#!/usr/bin/python3
#
# Copyright © 2021 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
#
# This software product is a proprietary product of Nvidia Corporation and its affiliates
# (the "Company") and all right, title, and interest in and to the software
# product, including all associated intellectual property rights, are and
# shall remain exclusively with the Company.
#
# This software product is governed by the End User License Agreement
# provided with the software product.
#
from __future__ import print_function
from enum import IntEnum

import sys
import argparse
import socket
import json
import re
import os

from ctypes import *
from ctypes import sizeof

g_debug_filter = 'INFO'

class ArgParseError(RuntimeError):
    ''' Argument parser error - runtime '''
    pass


class ChassisCtlCmd(IntEnum):
    ''' Chassis CTL command that will be sent via IPC to chassis manager'''
    MAINTENANCE_ENABLE = 1
    MAINTENANCE_DISABLE = 2
    LINECARD_ENABLE = 3
    LINECARD_DISABLE = 4
    SHOW_SUMMARY = 5
    SHOW_DETAIL = 6
    AUTO_PROVISION_ENABLE = 7
    AUTO_PROVISION_DISABLE = 8
    DEBUG_SLOT = 9
    INVALID = 10

class ChassisCtlCmdStatus(IntEnum):
    ''' Chassis CTL command status '''
    SUCCESS = 0
    INVALID_SLOT = 2
    INVALID_CMD = 3
    UNKNOWN_FAILURE = 4

class LineCardTypes(IntEnum):
    '''
    Linecards that are supported on SN4800; Got the following types from SDK
        - 4x400GE
        - 8x200GE
        - 16x100GE
    '''
    SX_MGMT_SLOT_CARD_TYPE_INVALID_E = 0
    SX_MGMT_SLOT_CARD_TYPE_MSN_4_400G_E = 1
    SX_MGMT_SLOT_CARD_TYPE_MSN_8_200G_E = 2
    SX_MGMT_SLOT_CARD_TYPE_MSN_16_100G_E = 3
    SX_MGMT_SLOT_CARD_TYPE_MIN = SX_MGMT_SLOT_CARD_TYPE_INVALID_E
    SX_MGMT_SLOT_CARD_TYPE_MAX = SX_MGMT_SLOT_CARD_TYPE_MSN_16_100G_E

class BaseStructure(Structure):
    '''
    This class is to intialize the ctypes/struct object.
    Any class that dereives from this class can set
    default values for the fields
    '''
    def __init__(self, **kwargs):
        """
        Ctypes.Structure with integrated default values.

        :param kwargs: values different to defaults
        :type kwargs: dict
        """

        values = type(self)._defaults_.copy()
        for (key, val) in kwargs.items():
            values[key] = val

        super().__init__(**values)            # Python 3 syntax

class ChassisCtlIpcCmd(BaseStructure):
    '''
    new IPC command header that will be sent over the wire; the server
    will send this header along with data
        cmd - IPC command type
        ipclen - Length of IPC command
        seq - Sequence number
        slot - Line card slot; It can be 1-8 or 255 (for all slots)
        status - response from server with information on whether the
               command succeeded or not
        type - provisioning type : It can be [16x100GE, 4x400GE]
    '''
    _pack_ = 1
    _fields_ = [("cmd", c_uint32),
                ("ipclen", c_uint32),
                ("seq", c_int32),
                ("slot", c_uint32),
                ("status", c_uint32),
                ("type", c_uint32),
               ]
    _defaults_ = {"cmd": ChassisCtlCmd.INVALID,
                  "ipclen": 0,
                  "seq": 0,
                  "slot": 0,
                  "status": 0,
                  "type": 0,
                 }

    def __new__(cls, in_cmd=None):
        if in_cmd is not None:
            return cls.from_buffer_copy(in_cmd)

        return Structure.__new__(cls)


class ChassisCtlIpcResp(BaseStructure):
    '''
    IPC response format that will be sent over the wire from
    chassis manager
    '''
    _pack_ = 1
    _fields_ = [("slot", c_uint32),
                ("serial", c_char * 32),
                ("card_type", c_char * 32),
                ("ini_major", c_uint32),
                ("ini_minor", c_uint32),
                ("state", c_char * 32),
                ("provisioned", c_uint32),
                ("uptime", c_uint64),
               ]
    _defaults_ = {"slot": 0,
                  "serial": b'',
                  "card_type": b'',
                  "ini_major": 0,
                  "ini_minor": 0,
                  "state": b'',
                  "provisioned": 0,
                  "uptime": 0
                 }

    def __new__(cls, in_resp=None):
        if in_resp is not None:
            return cls.from_buffer_copy(in_resp)

        return Structure.__new__(cls)

class ChassisCtlIpcDebugResp(BaseStructure):
    '''
    IPC response format that will be sent over the wire from
    chassis manager
    '''
    _pack_ = 1
    _fields_ = [("buffer", c_char * 256 * 1024 * 8)
               ]
    _defaults_ = {"buffer": b''}

    def __new__(cls, in_resp=None):
        if in_resp is not None:
            return cls.from_buffer_copy(in_resp)

        return Structure.__new__(cls)

class ChassisCtlIpcFailResp(BaseStructure):
    '''
    IPC resonse format when the command fails. The IPC command
    header will have status field which will indicate if the command
    succeeded/failed
    '''
    _pack_ = 1
    _fields_ = [("status", c_char * 128)]
    _defaults_ = {"status": b''}

    def __new__(cls, in_resp=None):
        if in_resp is not None:
            return cls.from_buffer_copy(in_resp)

        return Structure.__new__(cls)


def set_slot_msg(key, value, msg):
    '''
    Given key and value from the parse dictionary, set the slot
    appropriately in the CTL message; This API is called to build
    the CTL message
    '''
    if key == 'slot':
        if value == 'all':
            msg.slot = int(255)
        else:
            msg.slot = int(value)


def set_show_msg(key, value, msg):
    '''
    Given key and value from the parse dictionary, set the cmd
    appropriately in the CTL message; This API is called to build
    the CTL message
    '''
    if key == 'show':
        if value == 'summary':
            msg.cmd = ChassisCtlCmd.SHOW_SUMMARY
        elif value == 'detail':
            msg.cmd = ChassisCtlCmd.SHOW_DETAIL
        elif value == None:
            return None

    return 'success'


def set_provision_msg(key, value, msg):
    '''
    Given key and value from the parse dictionary, set the provision
    field appropriately in the CTL message; This API is called to build
    the CTL message
    '''
    if key == 'provision':
        if value == '16x100GE':
            msg.cmd = ChassisCtlCmd.AUTO_PROVISION_ENABLE
            msg.type = LineCardTypes.SX_MGMT_SLOT_CARD_TYPE_MSN_16_100G_E
        elif value == '4x400GE':
            msg.cmd = ChassisCtlCmd.AUTO_PROVISION_ENABLE
            msg.type = LineCardTypes.SX_MGMT_SLOT_CARD_TYPE_MSN_4_400G_E
        elif value == '8x200GE':
            msg.cmd = ChassisCtlCmd.AUTO_PROVISION_ENABLE
            msg.type = LineCardTypes.SX_MGMT_SLOT_CARD_TYPE_MSN_8_200G_E
        elif value == 'DISABLE':
            msg.cmd = ChassisCtlCmd.AUTO_PROVISION_DISABLE
            msg.type = LineCardTypes.SX_MGMT_SLOT_CARD_TYPE_INVALID_E
        else:
            return None

    return 'success'

def set_linecard_msg(key, value, msg):
    '''
    Given key and value from the parse dictionary, set the linecard
    field appropriately in the CTL message; This API is called to build
    the CTL message
    '''
    if key == 'linecard':
        if value == 'enable':
            msg.cmd = ChassisCtlCmd.LINECARD_ENABLE
        elif value == 'disable':
            msg.cmd = ChassisCtlCmd.LINECARD_DISABLE
        else:
            return None

    return 'success'


def set_maintenance_msg(key, value, msg):
    '''
    Given key and value from the parse dictionary, set the maintenance
    field appropriately in the CTL message; This API is called to build
    the CTL message
    '''
    if key == 'maintenance':
        if value == 'enable':
            msg.cmd = ChassisCtlCmd.MAINTENANCE_ENABLE
        elif value == 'disable':
            msg.cmd = ChassisCtlCmd.MAINTENANCE_DISABLE
        else:
            return None

    return 'success'

def set_debug_msg(key, value, msg):
    '''
    Given key and value from the parse dictionary, set the maintenance
    field appropriately in the CTL message; This API is called to build
    the CTL message
    '''
    if key == 'debug':
        if value == '1':
            msg.cmd = ChassisCtlCmd.DEBUG_SLOT
        else:
            return None

    return 'success'

def set_debug_filter(key, value):
    '''
    Given key and value from the parse dictionary, set the maintenance
    field appropriately in the CTL message; This API is called to build
    the CTL message
    '''
    global g_debug_filter
    if key == 'filter':
        g_debug_filter = value

    return 'success'

def build_ctl_msg(cmddict, seq):
    '''
    Given a dictionary of parsed cmdline, build CTL IPC message
    '''
    global g_debug_filter
    g_debug_filter = 'INFO'
    msg = ChassisCtlIpcCmd()
    msg.ipclen = sizeof(ChassisCtlIpcCmd)
    msg.seq = seq

    for key, value in cmddict.items():
        set_slot_msg(key, value, msg)

        if set_show_msg(key, value, msg) == None:
            continue

        if set_linecard_msg(key, value, msg) == None:
            continue

        if set_provision_msg(key, value, msg) == None:
            continue

        if set_maintenance_msg(key, value, msg) == None:
            continue

        if set_debug_msg(key, value, msg) == None:
            continue

        if set_debug_filter(key, value) == None:
            continue

    return msg


def parse_cmdline(argv):
    ''' Parse the command line '''
    parser = argparse.ArgumentParser(
        description='CTL interface to chassis manager')

    parser.add_argument('-l', '--slot',
                        dest="slot",
                        required=True,
                        action='store',
                        help='specify the linecard slot (\'all\' or 1-8).')
    parser.add_argument('-m', '--maintenance',
                        dest="maintenance",
                        required=False,
                        action='store',
                        help='maintenance mode enable/disable',
                        choices=['enable', 'disable'],
                        default='disable')
    parser.add_argument('-e', '--lcenable',
                        dest='linecard',
                        required=False,
                        action='store',
                        help='linecard dwmode enable/disable',
                        choices=['enable', 'disable'],
                        default='none')
    parser.add_argument('-f', '--gb-fw-update',
                        dest="gb-fw-path",
                        required=False,
                        action='store',
                        help='update linecard gearbox fw',
                        default=None)
    parser.add_argument('-p', '--gb-ini-update',
                        dest="gb-ini-path",
                        required=False,
                        action='store',
                        help='update linecard gearbox ini',
                        default=None)
    parser.add_argument('-s', '--show',
                        dest="show",
                        required=False,
                        action='store',
                        help='show detail/summary',
                        choices=['summary', 'detail'],
                        default=None)
    parser.add_argument('-c', '--provision',
                        dest="provision",
                        required=False,
                        action='store',
                        help='provision line card - provide type',
                        choices=['16x100GE', '4x400GE', 'DISABLE'],
                        default=None)
    parser.add_argument('-d', '--debug',
                        dest="debug",
                        required=False,
                        action='store_const',
                        const='1',
                        help='debug slot/s')
    parser.add_argument('-F', '--filter',
                        dest="filter",
                        required=False,
                        action='store',
                        choices=['INFO','WARN','ERR'],
                        help='debug slot/s')
    pretty_print = False
    parser.add_argument('-j', '--json',
                        dest="json",
                        required=False,
                        action='store_const',
                        help='display json output',
                        default=pretty_print,
                        const=not pretty_print)

    try:
        if len(argv) == 1:
            parser.print_help(sys.stderr)
            sys.exit(1)

        args = parser.parse_args(argv)
        mydict = vars(args)
        #print(mydict)
        return mydict
    except ArgParseError:
        exit(-1)


def send_ctl_toserver(cmd, cmddict):
    ''' connect to server and do IPC '''
    sock = socket.socket(socket.AF_UNIX,
                         socket.SOCK_STREAM)
    #print('connecting to server', file=sys.stderr)
    try:
        serveraddr = '/tmp/chassismgr'
        sock.connect(serveraddr)
        sock.sendall(cmd)

        if (cmd.cmd == ChassisCtlCmd.DEBUG_SLOT):
            data, sender = sock.recvfrom(256*1024*8)
        else:
            data, sender = sock.recvfrom(1024)
        #print('received ', len(data), file=sys.stderr)
        parse_ctl_response(cmd, data, cmddict, len(data))

    except socket.error as err:
        print('failed to connect', err, file=sys.stderr)
        sys.exit(1)

    finally:
        #print('closing socket', file=sys.stderr)
        sock.close()

def print_failure_bail(payload):
    '''
    We received IPC command failure; We need to decode the failure
    message in the response
    '''
    class Message(Structure):
        ''' a dynamic structure to unpack contents on the wire '''
        _pack_ = 1
        _fields_ = [("header", ChassisCtlIpcCmd),
                    ("resp", ChassisCtlIpcFailResp)
                   ]

        def __new__(cls, obj=None):
            if obj is not None:
                return cls.from_buffer_copy(obj)

            return Structure.__new__(cls)

    response = Message.from_buffer_copy(payload)
    print(response.resp.status.decode('utf-8'))
    return response


def get_ipcresponse_cnt(cmd, payload):
    '''
    Given the payload, find the number of IPC responses we received
    '''
    if sizeof(ChassisCtlIpcCmd) > len(payload):
        raise ValueError(-1, "Buffer smaller than header")

    header = ChassisCtlIpcCmd.from_buffer_copy(payload)
    ''' validate the response '''
    if validate_response(cmd, header) == False:
        print_failure_bail(payload)
        return int(0)

    if header.slot == 255:
        response_cnt = int(8)
    else:
        response_cnt = int(1)

    if response_cnt > len(payload):
        raise ValueError(-1, "Buffer smaller than header and data")

    return response_cnt


def validate_response(cmd, header):
    '''
    Validate the IPC command and the response
    '''
    if cmd.cmd != header.cmd:
        return False

    if header.status != ChassisCtlCmdStatus.SUCCESS:
        print('Validation failed with status %d' % header.status)
        return False

    return True


def parse_ctl_response(cmd, payload, cmddict, length=0):
    ''' API to parse the payload received on the wire'''
    response_cnt = get_ipcresponse_cnt(cmd, payload)
    #print("printing the response cnt = %d" % response_cnt)

    class Message(Structure):
        ''' a dynamic structure to unpack contents on the wire '''
        _pack_ = 1
        _fields_ = [("header", ChassisCtlIpcCmd),
                    ("response", ChassisCtlIpcResp * response_cnt)
                   ]

        def __new__(cls, obj=None):
            if obj is not None:
                return cls.from_buffer_copy(obj)

            return Structure.__new__(cls)

    class DebugMessage(Structure):
        ''' a dynamic structure to unpack contents on the wire '''
        _pack_ = 1
        if (length < sizeof(ChassisCtlIpcCmd)):
            assert "Packet length is not valid"
        else:
            _fields_ = [("header", ChassisCtlIpcCmd),
                        ("json_buffer", c_char * (length - sizeof(ChassisCtlIpcCmd)))
                        ]

        def __new__(cls, obj=None):
            if obj is not None:
                print('obj is not None')
                return cls.from_buffer_copy(obj)

            print('obj is None')
            return Structure.__new__(cls)

    if (cmd.cmd == ChassisCtlCmd.DEBUG_SLOT):
        message = DebugMessage.from_buffer_copy(payload)
    else:
        message = Message.from_buffer_copy(payload)

    print_ctl(message, response_cnt, cmddict)
    return message


def print_ctl(message, response_cnt, cmddict):
    ''' print CTL output on the console '''
    if message.header.cmd == ChassisCtlCmd.SHOW_SUMMARY:
        print_ctl_summary(message, response_cnt, cmddict)
    elif message.header.cmd == ChassisCtlCmd.SHOW_DETAIL:
        print_ctl_detail(message, response_cnt, cmddict)
    elif message.header.cmd == ChassisCtlCmd.DEBUG_SLOT:
        print_ctl_debug_slot(message, response_cnt, cmddict)

def print_json_summary(message, response_cnt, cmddict):
    '''API to print the summary in JSON format '''
    data = {}
    decoding = 'utf-8'
    for i in range(0, response_cnt):
        key = 'module' + str(message.response[i].slot)
        data[key] = {}
        data[key]['slot'] = message.response[i].slot
        data[key]['card_type'] = message.response[i].card_type.decode(decoding)
        data[key]['serial'] = message.response[i].serial.decode(decoding)
        data[key]['ini_major'] = message.response[i].ini_major
        data[key]['ini_minor'] = message.response[i].ini_minor
        data[key]['state'] = message.response[i].state.decode(decoding)

    json_data = json.dumps(data, indent=4)
    print(json_data)

def get_slot_state_bits(slot_state_str):
    slot_states = {'SLOT_PRESENT':'I', 'SLOT_VERIFIED':'V', 'SLOT_SYNCED':'S', 'SLOT_POWERED':'P', 'SLOT_ACTIVE':'A', 'SLOT_READY':'R'}
    ret_str = ''
    for key in slot_states:
        if key in slot_state_str:
            ret_str = ret_str + slot_states[key]
        else:
            ret_str = ret_str + 'X'
    
    return ret_str

def get_final_state(slot_state_str):
    slot_states = {'SLOT_PRESENT':'PRESENT', 'SLOT_VERIFIED':'VERIFIED', 'SLOT_SYNCED':'SYNCED', 'SLOT_POWERED':'POWERED', 'SLOT_ACTIVE':'ACTIVE', 'SLOT_READY':'READY'}
    ret_str = 'EMPTY'
    for key in slot_states:
        if key in slot_state_str:
            ret_str = slot_states[key]
    
    if ('ERR' in slot_state_str):
        ret_str = ret_str + '*'
    
    return ret_str   

def print_ctl_debug_slot(message, response_cnt, cmddict):
    global g_debug_filter
    decoding = 'utf-8'
    print_foot_note = False
    print("-" * 80)
    print('Chassis Details :\n')
    jsonResponse = json.loads(message.json_buffer.decode(decoding))
    slots = {'slot1','slot2','slot3','slot4','slot5','slot6','slot7','slot8'}
    is_online = lambda x : 'Yes' if ('SLOT_ACTIVE' in x) else ('No*' if ('ERR' in x) else 'No')
    config_state = lambda x : 'Disabled' if (x == 'True') else 'Enabled'
    tab = ''
    print('system uptime\t\t:\n\t' + os.popen('uptime\n').read().strip())

    syseeprom = os.popen('decode-syseeprom').read()
    platform_name = re.search('(.+)Platform Name(\s|\t)+[^\s\t]+\s+[^\s\t]+\s+(.+?)\n', syseeprom, re.DOTALL).group(3)
    print('Platform\t\t:\t%s' %platform_name)

    print('chassis serial no\t:\t' + os.popen('decode-syseeprom -e').read().strip())
    print("-" * 80)
    print('Slot Information: \n')
    print("Pre-provision:\tTrue - slot is pre-provisioned,\n\t\tFalse - slot is auto-provisioed")
    print('Config State :\tEnabled - slot will power up with card,\n\t\tDisabled - slot is administratively powered down')
    print("State Mask   :\tI - Inserted, V - Verified, S - Synced, P - Powered,\n\t\tA - Active, R - Ready")
    print('Slot Up-time is in hours:minutes:secs')
    print("-" * 80)
    #print('\x1b[1;30;47m' + 'slot|  Serial No  |   Type    |Pre-     |   Up-time  |Config  | State| HW      ' + '\x1b[0m')
    #print('\x1b[1;30;47m' + '    |             |           |Provision|            |State   | Mask | State   ' + '\x1b[0m')
    print('slot|  Serial No  |   Type    |Pre-     |   Up-time  |Config  | State| HW      ')
    print('    |             |           |Provision|            |State   | Mask | State   ')
    print("-" * 80)
    slot_list = {key:value for key, value in jsonResponse.items() if key in slots}
    for key, slot_str in slot_list.items():
        if 'Up Time' in slot_str:
            print("%2s   %-14s%-12s %-6s   %5d:%02d:%02d  %-5s  %s %-3s" % (slot_str['SLOT'], slot_str['Serial No'],
                    slot_str['Card Type'], 
                    slot_str["Provisioned"], slot_str['Up Time']['Hours'], slot_str['Up Time']['Minutes'],
                    slot_str['Up Time']['Seconds'], config_state(slot_str['Disabled']), get_slot_state_bits(slot_str['STATE']), 
                    get_final_state(slot_str['STATE'])))
        else:
            print("%2s   %-14s%-12s %-6s   %5d:%02d:%02d  %-5s  %s %-3s" % (slot_str['SLOT'], slot_str['Serial No'],
                    slot_str['Card Type'], 
                    slot_str["Provisioned"], 0, 0, 0,
                    config_state(slot_str['Disabled']), get_slot_state_bits(slot_str['STATE']),
                    get_final_state(slot_str['STATE'])))
            if ('*' in is_online(slot_str['STATE'])):
                print_foot_note = True
    if (print_foot_note == True):
        print("\n Note:\n    (*)Check log for state machine error")

    log_level = {'ERR': {'ERR'}, 'WARN' : {'ERR', 'WARN'}, 'INFO' : {'ERR', 'WARN', 'INFO'}}

    if not bool(g_debug_filter):
        return

    for key, value in slot_list.items():
        print('\x1b[4;37m' + '\nLogs for %s:' %(key) + '\x1b[0m')
        for log_key, log_string in value.items():
            if log_key == 'Log':
                for i in log_string:
                    if (re.search('\[(.+?):(.+?)\]', i).group(2) in log_level[g_debug_filter]):
                        print(i.strip())


def print_ctl_summary(message, response_cnt, cmddict):
    ''' print the CTL Summary on console '''
    if cmddict['json'] == True:
        print_json_summary(message, response_cnt, cmddict)
        return

    decoding = 'utf-8'
    print("-" * 50)
    print("{:<3} {:<18} {:<15}".format('Slot', 'Card type',
                                 'State'))
    print("-" * 50)
    for i in range(0, response_cnt):
        print("{:<3} {:<18} {:<15}".format(message.response[i].slot,
            message.response[i].card_type.decode(decoding),
            message.response[i].state.decode(decoding)))

        if response_cnt is 1:
            break



def print_ctl_detail(message, response_cnt, cmddict):
    ''' print the CTL Detail '''
    if cmddict['json'] == True:
        print_json_summary(message, response_cnt, cmddict)
        return

    decoding = 'utf-8'
    print("-" * 90)
    print("{:<3} {:<18} {:<15} {:<20} {:<20} {:<15}".format('Slot',
                                                            'Card type',
                                                            'Serial',
                                                            'INI major',
                                                            'INI minor',
                                                            'State'))
    print("-" * 90)
    for i in range(0, 8):
        print("{:<3} {:<18} {:<15} {:^20} {:^20} {:<15}".format(
            message.response[i].slot,
            message.response[i].card_type.decode(decoding),
            message.response[i].serial.decode(decoding),
            message.response[i].ini_major,
            message.response[i].ini_minor,
            message.response[i].state.decode(decoding)))

        if response_cnt is 1:
            break


