#!/usr/bin/env python
# Copyright (C) 2020-2021 NVIDIA Corporation. ALL RIGHTS RESERVED.
# Copyright 2016-2020 Cumulus Networks, Inc.
#
# mroute-check asic backend helper

import os
import re
import subprocess
import sys

try:
    from cumulus.mlx import mlx_open_connection
    from cumulus.mlx import mlx_close_connection
    from cumulus.mlx import mlx_get_mc_route
    from cumulus import mroute_check_util
except ImportError:
    raise NotImplementedError

class Kern:
    def __init__(self):
        self.vrf_get_db()

    def _get_cmd_output(self, cmdList, dispStr=None):
        try:
            with open(os.devnull, 'w') as fnull:
                dispStr = subprocess.check_output(cmdList, stderr=fnull)
        except subprocess.CalledProcessError:
            pass
        return dispStr

    def vrf_get_db(self):
        self.vrf_db = {}

        cmd = '/bin/ip vrf'
        dispStr = self._get_cmd_output(cmd.split(), '')

        for line in dispStr.splitlines():
            if 'Name              Table' not in line and '-----------------------' not in line:
                self.vrf_db[int(line.split()[1])] = line.split()[0]

    def vrf_get_str(self, table_id):
        return self.vrf_db[table_id]


class Sfs:
    #from switchd definitions:
    HAL_MROUTE_FLAG_VXLAN_ORIG = (1 << 2)
    HAL_MROUTE_FLAG_IN_HW = (1 << 6)
    MROUTE_DEFAULT_TABLE = 253

    def __init__(self):
        self.L3IntfSfsNode = "/cumulus/switchd/debug/l3_intf/info"
        self.MrouteSfsNode = "/cumulus/switchd/debug/mroute/info"
        self.switchd_debug_enabled = self.get_switchd_debug()
        if not self.switchd_debug_enabled:
            self.enable_switchd_debug()

        self.l3_intf_sfs_get()
        self.mroute_sfs_flags_get()

        if not self.switchd_debug_enabled:
            self.disable_switchd_debug()

    def l3_intf_sfs_get(self):
        l3_intf_raw_dump = {}
        self.l3_intf_db = {}

        fs = open(self.L3IntfSfsNode, "r")
        l3_intf_raw_dump = fs.read()
        fs.close()

        lines = l3_intf_raw_dump.splitlines()
        for l in lines:
            if "RIF" not in l:
                rif = l.split()[3]
                if rif == "vport":
                    rif = l.split()[5].replace('(', '').replace(')', '')
                if rif == "vlan":
                    rif = rif + l.split()[4]

                self.l3_intf_db[int(l.split()[0])] = rif

    def mroute_sfs_flags_get(self):
        mroute_raw_dump = {}
        self.mroute_flags_db = {}

        fs = open(self.MrouteSfsNode, "r")
        mroute_raw_dump = fs.read()
        fs.close()

        kern = Kern()

        lines = mroute_raw_dump.splitlines()
        for l in lines:

            #Pattern of SFS mroute line looks like
            #       Mroute: tbl 253 (0.0.0.0, 239.1.1.106) iif (uplink-2), pcnt 3, flags 0x9
            if "Mroute" in l:
                l = l.replace('(', '').replace(')', '').replace(',', '')
                pat = re.compile('\s*Mroute: tbl (?P<tbl>\S+) (?P<src>\S+) (?P<grp>\S+) iif (?P<iif>\S+) pcnt (?P<pcnt>\S+) flags (?P<flags>\S+).*')
                obj = pat.match(l)

                table_id = int(obj.group('tbl'))
                if table_id == self.MROUTE_DEFAULT_TABLE:
                    vrf = "default"
                else:
                    vrf = kern.vrf_get_str(table_id)

                src = obj.group('src')
                grp = obj.group('grp')
                flags = obj.group('flags')

                self.mroute_flags_db['%s, (%s,%s)' % (vrf, src, grp)] = flags

    def mroute_is_hw_installed(self, key, vrf, src, grp):
        if key is None:
            flags = self.mroute_flags_db['%s, (%s,%s)' % (vrf, src, grp)]
        else:
            flags = self.mroute_flags_db[key]

        flags = int(flags, base=16)
        if flags & self.HAL_MROUTE_FLAG_IN_HW:
            return True

        return False

    def mroute_is_ulm_orig(self, key, vrf, src, grp):
        if key is None:
            flags = self.mroute_flags_db['%s, (%s,%s)' % (vrf, src, grp)]
        else:
            flags = self.mroute_flags_db[key]

        flags = int(flags, base=16)
        if flags & self.HAL_MROUTE_FLAG_VXLAN_ORIG:
            return True

        return False

    def get_switchd_debug(self):
        return os.path.isfile(self.L3IntfSfsNode)

    def enable_switchd_debug(self):
        try:
            os.system('echo 1 > /cumulus/switchd/ctrl/debug')
        except:
            raise Exception("Switchd Fuse Node not enabled")

    def disable_switchd_debug(self):
        try:
            os.system('echo 1 > /cumulus/switchd/ctrl/debug')
        except:
            raise Exception("Switchd Fuse Node not disabled")

    def l3IntfStr(self, IntfId):
         return (self.l3_intf_db[IntfId])

class Hmroute:
    #loopback interface is set as dont care RIF in SDK
    DONT_CARE_IIF_STR = "lo"

    #From SDK definitions:
    SP1_DONT_CARE_IIF = 1001
    SP2_DONT_CARE_IIF = 4001
    MLX_TUNNEL_ID = 0x8c00000

    def __init__(self, sip, dip, iif, erifs, vrf, internal_mroute):
        self.vrf = vrf
        self.sip = sip
        self.dip = dip
        self.erifs = []
        self.mismatch = []
        self.is_ulm_term = 0
        self.is_ulm_orig = 0
        self.internal_iif = 0
        self.internal_erifs = []

        self.sfs = Sfs()

        self.iif = self.iifStr(iif)

        for eriflist in erifs:
            for erif in eriflist:
                if erif == self.MLX_TUNNEL_ID:
                    self.is_ulm_term = 1
                else:
                    self.erifs.append(self.sfs.l3IntfStr(erif))

        if internal_mroute:
            self.internal_iif = self.iifStr(internal_mroute['iif'])

            internal_erifs = internal_mroute['egress_rifs']
            for eriflist in internal_erifs:
                for erif in eriflist:
                    if erif == self.MLX_TUNNEL_ID:
                        self.is_ulm_term = 1
                    else:
                        self.internal_erifs.append(self.sfs.l3IntfStr(erif))

        if self.sfs.mroute_is_ulm_orig(None, vrf, sip, dip):
            self.is_ulm_orig = 1

    def iifStr(self, iif):
        if iif in [ self.SP1_DONT_CARE_IIF, self.SP2_DONT_CARE_IIF ]:
            return self.DONT_CARE_IIF_STR
        else:
            return self.sfs.l3IntfStr(iif)

    def __str__(self):
        mroute_s_g_str = '(%s,%s)' % (self.sip, self.dip)
        mroute_str = '%-*sIif: %s Oifs:' % (33, mroute_s_g_str, self.iif)
        for erif in self.erifs:
            mroute_str += " "
            mroute_str += erif

        mroute_str += " Table: %s" % (self.vrf)
        if self.is_ulm_orig:
            mroute_str += " Flags: evpn-pim-orig"
        if self.is_ulm_term:
            mroute_str += " Flags: evpn-pim-term"

        if self.internal_iif:
            mroute_str += '\n%-*sIif: %s Oifs:' % (33, mroute_s_g_str, self.internal_iif)
            for erif in self.internal_erifs:
                mroute_str += " "
                mroute_str += erif

            mroute_str += " Table: %s" % (self.vrf)
            mroute_str += " Flags: Internal"

        if self.mismatch:
            mroute_str += '\n  errors:'
        for mismatch in self.mismatch:
            mroute_str += '\n    %s' % mismatch

        mroute_str += '\n'

        return mroute_str

class Asic:
    ORIG_PEERLINK_DEV = "peerlink-3.4094"
    ORIG_LOOPBACK_DEV = "lo"
    TERM_DEV = "ipmr-lo"

    def __init__(self):
        # key '(a.b.c.d, a.b.c.d)'
        self.desc = "asic"
        self.vrf_db = mroute_check_util.VrfDb()

        mlx_open_connection()
        self.mroutes = {}
        self._get_mlx_mroute()
        mlx_close_connection()

    def _get_mlx_mroute(self):
        """
        mcroute_table format:
        0.0.0.0/239.1.1.3 {'egress_rifs': [36L, 4L], 'iif': 5L, 'vrid': 0, 'egress_rif_cnt': 2}
        """
        mcroute_table = mlx_get_mc_route()
        for mc in mcroute_table.items():
            src, grp = mc[0].split()[1].split('/')
            erifs = []
            vrf = self.vrf_db.vrf_id2name(int(mc[1].get('vrid')))
            iif = mc[1].get('iif')
            erifs = mc[1].get('egress_rifs')
            internal_mroute = mc[1].get('internal_mroute')
            mroute = Hmroute(src, grp, iif, erifs, vrf, internal_mroute)
            self.mroutes['%s, (%s,%s)' % (vrf, src, grp)] = mroute

    def __str__(self):
        ipmr_str = 'Asic ip mroutes:\n'
        for key in sorted(self.mroutes):
            ipmr_str += self._dump_one_mroute(self.mroutes[key])
        return ipmr_str

    def _dump_one_mroute(self, mroute):
        mroute_str = '%s' % (mroute)
        return mroute_str

    def mroute_eq(self, key, iif, oil, ul_mcast):
        self.sfs = Sfs()

        if self.sfs.mroute_is_hw_installed(key, None, None, None) != True:
            # mroute is not installed in HW because of OIL with pimreg/iif with lo
            return True

        mroute = self.mroutes.get(key)
        if not mroute:
            return False

        if mroute.iif != iif:
            mroute.mismatch.append('kernel IIF %s does not match asic IIF %s' % (iif, mroute.iif))

        hw_oil = set(mroute.erifs)
        k_oil = set(oil)

        # the kernel requires that iif be partof oil for (*,G)
        if iif in k_oil:
            k_oil.remove(iif)

        if mroute.is_ulm_term and mroute.is_ulm_orig:
            mroute.mismatch.append('Mroute marked for both Vxlan origination and termination')

        if self.TERM_DEV in k_oil:
            if mroute.is_ulm_term == 0:
                mroute.mismatch.append('HW not setup for multicast tunnel decapsulation')
            k_oil.remove(self.TERM_DEV)

        diff = hw_oil - k_oil
        if diff:
             mroute.mismatch.append('HW has OIFs %s not in the kernel' % (diff))

        diff = k_oil - hw_oil
        if diff:
            mroute.mismatch.append('kernel has OIFs %s not in HW' % (diff))

        if mroute.is_ulm_orig and iif == self.ORIG_PEERLINK_DEV:
            if mroute.internal_iif != self.ORIG_LOOPBACK_DEV:
                 mroute.mismatch.append('Internal mroute IIF incorrect')

            hw_internal_oil = set(mroute.internal_erifs)
            k_internal_oil = set(oil)

            diff = hw_internal_oil - k_internal_oil
            if diff:
                mroute.mismatch.append('Internal mroute HW has OIFs %s not in the kernel' % (diff))

            diff = k_internal_oil - hw_internal_oil
            if diff:
                mroute.mismatch.append('Internal mroute HW doesnt have OIFs %s found in kernel' % (diff))

        if mroute.mismatch:
            return False

        return True

    def get_mroutes(self):
        # return mroute keys as a set
        return set(self.mroutes)

    def dump_mroutes(self, filter_mroutes=None):
        if not filter_mroutes:
            for key in sorted(self.mroutes):
                sys.stdout.write('%s\n' % self._dump_one_mroute(self.mroutes[key]))
            return

        for key in filter_mroutes:
            mroute = self.mroutes.get(key)
            if mroute:
                sys.stdout.write('%s\n' % self._dump_one_mroute(self.mroutes[key]))

    def get_mroute_iif_and_oil(self, key):
        mroute = self.mroutes.get(key)
        if mroute:
            return mroute.iif, mroute.erifs
        return None, None
